import os
import uuid
from langchain_community.document_loaders import DirectoryLoader, TextLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_ollama import OllamaEmbeddings
from langchain_community.vectorstores import Chroma
from langchain_community.llms import Ollama
from langchain.chains import RetrievalQA
from langchain.prompts import PromptTemplate

from code_rag.doc_tracker import DocumentTracker


class RAG:
    def __init__(self, docs_dir, db_dir, tracker_file):
        self.docs_dir = docs_dir
        self.db_dir = db_dir
        self.tracker = DocumentTracker(tracker_file)

    def process_documents(self, files, text_splitter):
        """Process document files into chunks with tracking metadata"""
        all_chunks = []
        file_chunk_map = {}

        for file_path in files:
            # Load the document
            loader = TextLoader(file_path)
            documents = loader.load()

            # Add source metadata
            for doc in documents:
                doc.metadata["source"] = file_path
                doc.metadata["source_id"] = file_path  # For easier identification

            # Split the document
            chunks = text_splitter.split_documents(documents)

            # Generate and track chunk IDs
            chunk_ids = []
            for chunk in chunks:
                chunk_id = str(uuid.uuid4())
                chunk.metadata["chunk_id"] = chunk_id
                chunk_ids.append(chunk_id)

            # Store chunk mappings if tracker is provided
            if self.tracker:
                self.tracker.update_chunk_mappings(file_path, chunk_ids)

            file_chunk_map[file_path] = chunks
            all_chunks.extend(chunks)

        return all_chunks, file_chunk_map

    def create_vector_db(self, extension=".txt", force_refresh=False):
        """
        Create or update a vector database, with complete handling of changes
        """
        # Text splitter configuration
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000, chunk_overlap=200
        )

        # Create embeddings
        embeddings = OllamaEmbeddings(model="nomic-embed-text")

        if force_refresh:
            print("Force refresh: Processing all documents")
            # Load all documents
            loader = DirectoryLoader(self.docs_dir, glob=f"**/*{extension}")
            all_documents = loader.load()

            # Add unique IDs to each document
            for doc in all_documents:
                doc.metadata["source"] = os.path.abspath(doc.metadata["source"])
                doc.metadata["source_id"] = doc.metadata["source"]

            # Split documents
            chunks = text_splitter.split_documents(all_documents)

            # Add chunk IDs and update tracker
            file_chunk_map = {}
            for chunk in chunks:
                chunk_id = str(uuid.uuid4())
                chunk.metadata["chunk_id"] = chunk_id

                source = chunk.metadata["source"]
                if source not in file_chunk_map:
                    file_chunk_map[source] = []

                file_chunk_map[source].append(chunk_id)

            # Update tracker with chunk mappings
            for file_path, chunk_ids in file_chunk_map.items():
                self.tracker.update_chunk_mappings(file_path, chunk_ids)

            print(
                f"Processing {len(all_documents)} documents with {len(chunks)} chunks"
            )

            # Create new vector store
            vectorstore = Chroma.from_documents(
                documents=chunks, embedding=embeddings, persist_directory=self.db_dir
            )
            vectorstore.persist()
            print(f"Created new vector database at {self.db_dir}")

            return vectorstore

        # Get changes since last update
        changed_files = self.tracker.get_changed_files(self.docs_dir)

        if not any(changed_files.values()):
            print("No document changes detected")
            # Load existing vector store if available
            if os.path.exists(self.db_dir):
                return Chroma(
                    persist_directory=self.db_dir, embedding_function=embeddings
                )
            else:
                print("No vector database exists. Creating from all documents...")
                return self.create_vector_db(force_refresh=True)

        # Process changes
        print(
            f"Changes detected - New: {len(changed_files['new'])}, Modified: {len(changed_files['modified'])}, Deleted: {len(changed_files['deleted'])}"
        )

        # Load existing vector store if it exists
        if os.path.exists(self.db_dir):
            vectorstore = Chroma(
                persist_directory=self.db_dir, embedding_function=embeddings
            )

            # 1. Handle deleted documents
            if changed_files["deleted"]:
                chunks_to_delete = self.tracker.get_chunks_to_delete(
                    changed_files["deleted"]
                )
                if chunks_to_delete:
                    print(
                        f"Removing {len(chunks_to_delete)} chunks from deleted documents"
                    )
                    # Delete the chunks from vector store
                    vectorstore._collection.delete(
                        where={"chunk_id": {"$in": chunks_to_delete}}
                    )

            # 2. Handle modified documents (delete old chunks first)
            chunks_to_delete_modified = self.tracker.get_chunks_for_modified_files(
                changed_files["modified"]
            )
            if chunks_to_delete_modified:
                print(
                    f"Removing {len(chunks_to_delete_modified)} chunks from modified documents"
                )
                vectorstore._collection.delete(
                    where={"chunk_id": {"$in": chunks_to_delete_modified}}
                )

            # 3. Process new and modified documents
            files_to_process = changed_files["new"] + changed_files["modified"]
            if files_to_process:
                chunks, _ = self.process_documents(
                    files_to_process, text_splitter, self.tracker
                )
                print(f"Adding {len(chunks)} new chunks to the vector store")
                vectorstore.add_documents(chunks)
        else:
            # If no existing DB, create from all documents
            print("No existing vector database. Creating from all documents...")
            return self.create_vector_db(force_refresh=True)

        # Persist changes
        vectorstore.persist()
        print(f"Vector database updated at {self.db_dir}")

        return vectorstore

    def setup_rag(self, model_name="llama3"):
        """
        Set up the RAG system with an existing vector database
        """
        # Load the embeddings
        embeddings = OllamaEmbeddings(model="nomic-embed-text")

        # Load the vector store
        vectorstore = Chroma(
            persist_directory=self.db_dir, embedding_function=embeddings
        )

        # Create a retriever
        retriever = vectorstore.as_retriever(search_kwargs={"k": 4})

        # Set up the LLM
        llm = Ollama(model=model_name)

        # Create a custom prompt template
        template = """
        Answer the question based on the context provided. If you don't know the answer,
        just say you don't know. Don't try to make up an answer.

        Context: {context}

        Question: {question}

        Answer:
        """

        prompt = PromptTemplate(
            input_variables=["context", "question"], template=template
        )

        # Create the RAG chain
        rag_chain = RetrievalQA.from_chain_type(
            llm=llm,
            chain_type="stuff",
            retriever=retriever,
            chain_type_kwargs={"prompt": prompt},
        )

        return rag_chain

    def query_rag(self, rag_chain, query):
        """
        Query the RAG system
        """
        response = rag_chain.invoke({"query": query})
        return response["result"]
